/* AIPlayer.c   Code for dopewars computer players                      */
/* Copyright (C)  1998-9  Ben Webb                                      */
/*                Email: ben@bellatrix.pcl.ox.ac.uk                     */
/*                WWW: http://bellatrix.pcl.ox.ac.uk/~ben/dopewars/     */

/* This program is free software; you can redistribute it and/or        */
/* modify it under the terms of the GNU General Public License          */
/* as published by the Free Software Foundation; either version 2       */
/* of the License, or (at your option) any later version.               */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program; if not, write to the Free Software          */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               */
/*                   MA  02111-1307, USA.                               */


#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>
#include "dopewars.h"
#include "message.h"
#include "clientside.h"
#include "AIPlayer.h"

#define NUMNAMES      3
#define MINSAFECASH   300
#define TURNPAUSE     5
#define MINSAFEHEALTH 40

/* Reserve some space for picking up new guns */
#define SPACERESERVE  10

/* Suitable names for AI players. Chosen at random. */
char *Names[NUMNAMES] = {
   "Chip", "Dopey", "Al" 
};

void AIPlayerLoop() {
/* Main loop for AI players. Connects to server, plays game, */
/* and then disconnects.                                     */
   char *pt,text[BUFLEN];
   struct PLAYER *AIPlay;
   fd_set readfs;
   AIPlay=AddPlayer(0,&FirstClient);
   if (!AIPlay) {
      printf("Error: Cannot allocate memory. AI Player terminating \
abnormally.\n");
      return;
   }
   printf("AI Player started; attempting to contact server ");
   printf("at %s:%d...\n",ServerName,Port);
   pt=SetupNetwork();
   if (pt) {
      printf("Error: Could not connect to dopewars server\n(%s)\n",pt);
      printf("AI Player terminating abnormally.\n");
      return;
   }
   AISetName(AIPlay);
   printf("Connection established\n");

   while (1) {
      FD_ZERO(&readfs);
      FD_SET(ClientSock,&readfs);
      if (select(ClientSock+1,&readfs,NULL,NULL,NULL)==-1) {
         if (errno==EINTR) continue;
         printf("Error in select\n"); exit(1);
      }
      if (FD_ISSET(ClientSock,&readfs)) {
         pt=bgets(text,BUFLEN,ClientSock);
         chomp(text);
         if (!pt) {
            printf("Connection to server lost!\n");
            ShutdownNetwork();
            break;
         } else {
            if (HandleAIMessage(text,AIPlay)) {
               ShutdownNetwork();
               break;
            }
         }
      }
   }
   sleep(1);
   printf("AI Player terminated OK.\n");
}

void AISetName(struct PLAYER *AIPlay) {
/* Chooses a random name for the AI player, and informs the server */
   sprintf(AIPlay->Name,"AI) %s",Names[brandom(0,NUMNAMES)]);
   SendClientMessage(NULL,C_NONE,C_NAME,NULL,AIPlay->Name);
   printf("Using name %s\n",AIPlay->Name);
}

int HandleAIMessage(char *Message,struct PLAYER *AIPlay) {
/* Performs appropriate processing on an incoming network message */
/* "Message" for AI player "AIPlay". Returns 1 if the game should */
/* be ended as a result, 0 otherwise.                             */
   char Data[BUFLEN],Code,AICode,WasFighting;
   struct PLAYER *From,*To,*tmp;
   if (ProcessMessage(Message,&From,&AICode,&Code,&To,Data,FirstClient)==-1) {
      printf("Bad network message. Oops.\n"); return 0;
   }
   switch(Code) {
      case C_LIST:
         tmp=AddPlayer(0,&FirstClient);
         strcpy(tmp->Name,Data);
         break;
      case C_ENDLIST:
         printf("Players in this game:-\n");
         for (tmp=FirstClient;tmp;tmp=tmp->Next) {
            printf("    %s\n",tmp->Name);
         }
         break;
      case C_NEWNAME:
         AISetName(AIPlay);
         break;
      case C_PRINTMESSAGE:
         PrintAIMessage(Data);
         break;
      case C_MSG:
         printf("%s: %s\n",From->Name,Data);
         break;
      case C_MSGTO:
         printf("%s->%s: %s\n",From->Name,To->Name,Data);
         break;
      case C_JOIN:
         printf("%s joins the game.\n",Data);
         tmp=AddPlayer(0,&FirstClient);
         strcpy(tmp->Name,Data);
         break;
      case C_LEAVE:
         if (From!=&Noone) {
            printf("%s has left the game.\n",From->Name);
            RemovePlayer(From,&FirstClient);
         }
         break;
      case C_SUBWAYFLASH:
         sleep(TURNPAUSE);
         printf("Jetting to %s with %ld cash and %ld debt\n",
                Location[(int)AIPlay->IsAt].Name,AIPlay->Cash,AIPlay->Debt);
         if (brandom(0,100)<10) AISendRandomMessage(AIPlay);
         break;
      case C_UPDATE:
         WasFighting=FALSE;
         if (From==&Noone) {
            if (AIPlay->Flags & FIGHTING) WasFighting=TRUE;
            ReceivePlayerData(Data,To);
         } else {
            ReceivePlayerData(Data,From); /* spy reports */
         }
         if (!(AIPlay->Flags & FIGHTING) && WasFighting) {
            AIDealDrugs(AIPlay);
            AIJet(AIPlay);
         }
         if (AIPlay->Health==0) {
            printf("AI Player killed. Terminating normally.\n");
           return 1;
         }
         break;
      case C_DRUGHERE:
         ReceiveDrugsHere(Data,To);
         AIDealDrugs(AIPlay);
         AIJet(AIPlay);
         break;
      case C_GUNSHOP:
         AIGunShop(AIPlay);
         break;
      case C_LOANSHARK:
         AIPayLoan(AIPlay);
         break;
      case C_QUESTION:
         AIHandleQuestion(Data,AICode,AIPlay,From);
         break;
      case C_SAYING: case C_HISCORE: case C_STARTHISCORE:
         break;
      case C_ENDHISCORE:
         printf("Game time is up. Leaving game.\n");
         return 1;
      default:
         printf("%s^%c^%s%s\n",From->Name,Code,To->Name,Data);
         break;
   }
   return 0;
}

void PrintAIMessage(char *Text) {
/* Prints a message received via a printmessage or question */
/* network message, stored in "Text"                        */
   int i;
   char SomeText=0;
   for (i=0;i<strlen(Text);i++) {
      if (Text[i]=='^') {
         if (SomeText) putchar('\n'); 
      } else {
         putchar(Text[i]);
         SomeText=1;
      }
   }
   putchar('\n');
}

void AIDealDrugs(struct PLAYER *AIPlay) {
/* Buy and sell drugs for AI player "AIPlay" */
   long Profit[NUMDRUG],MaxProfit;
   char text[BUFLEN];
   int i,LastHighest,Highest,Num,MinProfit;
   for (i=0;i<NUMDRUG;i++) {
      Profit[i]=AIPlay->Here->Price[i]-(Drug[i].MaxPrice+Drug[i].MinPrice)/2;
   }
   MinProfit=0;
   for (i=0;i<NUMDRUG;i++) if (Profit[i]<MinProfit) MinProfit=Profit[i];
   MinProfit--;
   for (i=0;i<NUMDRUG;i++) if (Profit[i]<0) Profit[i]=MinProfit-Profit[i];
   LastHighest=-1;
   while (1) {
      MaxProfit=MinProfit;
      Highest=-1;
      for (i=0;i<NUMDRUG;i++) {
         if (Profit[i]>MaxProfit && i!=LastHighest && 
             (LastHighest==-1 || Profit[LastHighest]>Profit[i])) {
            Highest=i;
            MaxProfit=Profit[i];
         }
      }
      LastHighest=Highest;
      if (Highest==-1) break;
      Num=AIPlay->Carry.Drug[Highest];
      if (MaxProfit>0 && Num>0) {
         printf("Selling %d %s at $%ld\n",Num,Drug[Highest].Name,
                AIPlay->Here->Price[Highest]);
         AIPlay->CoatSize+=Num;
         AIPlay->Cash+=Num*AIPlay->Here->Price[Highest];
         sprintf(text,"drug^%d^%d",Highest,-Num);
         SendClientMessage(AIPlay,C_NONE,C_BUYOBJECT,NULL,text);
      }
      if (AIPlay->Here->Price[Highest] != 0 &&
          AIPlay->CoatSize>SPACERESERVE) {
         Num=AIPlay->Cash/AIPlay->Here->Price[Highest];
         if (Num>AIPlay->CoatSize-SPACERESERVE) {
            Num=AIPlay->CoatSize-SPACERESERVE;
         }
         if (MaxProfit<0 && Num>0) {
            printf("Buying %d %s at $%ld\n",Num,Drug[Highest].Name,
                   AIPlay->Here->Price[Highest]);
            sprintf(text,"drug^%d^%d",Highest,Num);
            AIPlay->CoatSize-=Num;
            AIPlay->Cash-=Num*AIPlay->Here->Price[Highest];
            SendClientMessage(AIPlay,C_NONE,C_BUYOBJECT,NULL,text);
         }
      }
   }
}

void AIGunShop(struct PLAYER *AIPlay) {
/* Handles a visit to the gun shop by AI player "AIPlay" */
   SendClientMessage(AIPlay,C_NONE,C_DONE,NULL,NULL);
}

void AIJet(struct PLAYER *AIPlay) {
/* Decides on a new game location for AI player "AIPlay" and jets there */
   int NewLocation;
   char text[40];
   if (!AIPlay) return;
   NewLocation=AIPlay->IsAt;
   if (AIPlay->Cash > (long)((float)AIPlay->Debt*1.2)) NewLocation=LOANSHARK;
   while (NewLocation==AIPlay->IsAt) NewLocation=brandom(0,NUMLOCATION);
   sprintf(text,"%d",NewLocation);
   SendClientMessage(AIPlay,C_NONE,C_REQUESTJET,NULL,text);
}

void AIPayLoan(struct PLAYER *AIPlay) {
/* Pays off the loan of AI player "AIPlay" if this doesn't leave */
/* the player with insufficient funds for further dealing        */
   char text[BUFLEN];
   if (AIPlay->Cash-AIPlay->Debt >= MINSAFECASH) {
      sprintf(text,"%ld",AIPlay->Debt);
      printf("Debt of %ld paid off to loan shark\n",AIPlay->Debt);
      SendClientMessage(AIPlay,C_NONE,C_PAYLOAN,NULL,text);
   }
   SendClientMessage(AIPlay,C_NONE,C_DONE,NULL,NULL);
}

void AISendAnswer(struct PLAYER *From,struct PLAYER *To,char *answer) {
/* Sends the answer "answer" from AI player "From" to the server,        */
/* claiming to be for player "To". Also prints the answer on the screen. */
   SendClientMessage(From,C_NONE,C_ANSWER,To,answer); puts(answer);
}

void AIHandleQuestion(char *Data,char AICode,struct PLAYER *AIPlay,
     struct PLAYER *From) {
/* Works out a sensible response to the question coded in "Data" and with */
/* computer-readable code "AICode", claiming to be from "From" and for AI */
/* player "AIPlay", and sends it                                          */
   char *Prompt,allowed[BUFLEN];
   if (From==&Noone) From=NULL;
   Prompt=Data;
   Prompt=ExtractWord(allowed,Prompt);
   PrintAIMessage(Prompt);
   switch (AICode) {
      case C_ASKLOAN: case C_ASKGUNSHOP: case C_ASKPUB:
      case C_ASKBITCH: case C_ASKRUN: case C_ASKGUN:
         AISendAnswer(AIPlay,From,"Y");
         break;
      case C_ASKRUNFIGHT:
         AISendAnswer(AIPlay,From,AIPlay->Health<MINSAFEHEALTH ? "R" : "F");
         break;
      case C_ASKBANK:
         AISendAnswer(AIPlay,From,"N");
         break;
      case C_ATTACKEVADE:
         AISendAnswer(AIPlay,From,"A");
         break;
      case C_ASKSEW:
         AISendAnswer(AIPlay,From,AIPlay->Health<MINSAFEHEALTH ? "Y" : "N");
         break;
      default:
         AISendAnswer(AIPlay,From,"N");
         break;
   }
}

void AISendRandomMessage(struct PLAYER *AIPlay) {
/* Sends a random message to all other dopewars players */
   char *RandomInsult[5]= {
      "Call yourselves drug dealers?",
      "A trained monkey could do better...",
      "Think you\'re hard enough to deal with the likes of me?",
      "Zzzzz... are you dealing in candy or what?",
      "Reckon I'll just have to shoot you for your own good."
   };
   SendClientMessage(AIPlay,C_NONE,C_MSG,NULL,RandomInsult[brandom(0,5)]);
}
