/* clientside.c   Handles the client side of Dopewars                   */
/* Copyright (C)  1998-9  Ben Webb                                      */
/*                Email: ben@bellatrix.pcl.ox.ac.uk                     */
/*                WWW: http://bellatrix.pcl.ox.ac.uk/~ben/dopewars/     */

/* This program is free software; you can redistribute it and/or        */
/* modify it under the terms of the GNU General Public License          */
/* as published by the Free Software Foundation; either version 2       */
/* of the License, or (at your option) any later version.               */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program; if not, write to the Free Software          */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston,               */
/*                   MA  02111-1307, USA.                               */


#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <ctype.h>
#include "clientside.h"
#include "serverside.h"
#include "dopewars.h"
#include "message.h"
#include "AIPlayer.h"

struct PLAYER *FirstClient;
void PrepareHighScoreScreen();
void PrintHighScore(char *Data);

int HandleClientMessage(char *Message,struct PLAYER *ReallyTo) {
/* Given a message "Message" coming in on a socket which identifies it as   */
/* "really" for player "ReallyTo", performs processing and reacts properly; */
/* returns 0 on normal termination, or 1 if the message indicates that the  */
/* client should quit.                                                      */
   char *pt,Data[BUFLEN],Code,wrd[BUFLEN],buf[BUFLEN];
   char AICode;
   struct PLAYER *From,*To,*tmp;
   int i;
   struct timeval tv;
   if (ProcessMessage(Message,&From,&AICode,&Code,&To,Data,FirstClient)==-1) {
      mvaddstr(22,0,"Bad network message. Oops. ");
      nice_wait(); return 0;
   }

   switch(Code) {
      case C_LIST:
         tmp=AddPlayer(0,&FirstClient);
         strcpy(tmp->Name,Data);
         break;
      case C_ENDLIST:
         if (FirstClient && FirstClient->Next) ListPlayers(To,FALSE,NULL);
         break;
      case C_STARTHISCORE:
         PrepareHighScoreScreen();
         break;
      case C_HISCORE:
         PrintHighScore(Data);
         break;
      case C_ENDHISCORE:
         return 1;
      case C_PUSH:
         mvaddstr(22,0,"You have been pushed from the server.");
         ShutdownNetwork();
         nice_wait();
         break;
      case C_QUIT:
         mvaddstr(22,0,"The server has terminated.");
         ShutdownNetwork();
         nice_wait();
         break;
      case C_MSG:
         sprintf(buf,"%s: %s",From->Name,Data);
         display_message(buf);
         break;
      case C_MSGTO:
         sprintf(buf,"%s->%s: %s",From->Name,To->Name,Data);
         display_message(buf);
         break;
      case C_JOIN:
         sprintf(buf,"%s breezes into town!",Data);
         mvaddstr(22,0,buf); nice_wait();
         tmp=AddPlayer(0,&FirstClient);
         strcpy(tmp->Name,Data);
         break;
      case C_LEAVE:
         if (From!=&Noone) {
            sprintf(buf,"%s has left town.",From->Name);
            RemovePlayer(From,&FirstClient);
            mvaddstr(22,0,buf); nice_wait();
         }
         break;
      case C_RENAME:
         sprintf(buf,"%s will now be known as %s.",From->Name,Data);
         strcpy(From->Name,Data);
         mvaddstr(22,0,buf); nice_wait();
         break;
      case C_PRINTMESSAGE:
         PrintMessage(Data);
         nice_wait();
         break;
      case C_DRUGHERE:
         ReceiveDrugsHere(Data,To);
         break;
      case C_SUBWAYFLASH:
         for (i=0;i<4;i++) {
            print_location("S U B W A Y");
            refresh();
            tv.tv_sec=0;
            tv.tv_usec=100000;
            select(0,NULL,NULL,NULL,&tv);
            print_location("");
            refresh();
            tv.tv_sec=0;
            tv.tv_usec=100000;
            select(0,NULL,NULL,NULL,&tv);
         }
         print_location(Location[(int)To->IsAt].Name);
         break;
      case C_SAYING:
         attrset(TextAttr);
         clear_bottom();
         if (brandom(0,100)<50) {
            mvaddstr(17,1,"The lady next to you on the subway said,");
            mvaddch(18,3,'\"');
            addstr(SubwaySaying[brandom(0,NUMSUBWAY)]);
            addch('\"');
            if (brandom(0,100)>70) {
               mvaddstr(19,4,"(at least, you -think- that's what she said)");
            }
         } else {
            mvaddstr(17,1,"You hear someone playing ");
            addstr(Playing[brandom(0,NUMPLAYING)]);
         }
         nice_wait();
         break;
      case C_QUESTION:
         pt=Data;
         pt=ExtractWord(wrd,pt);
         PrintMessage(pt);
         addch(' ');
         i=GetKey(wrd);
         buf[0]=i; buf[1]=0;
         SendClientMessage(To,C_NONE,C_ANSWER,From==&Noone ? NULL : From,buf);
         break;
      case C_LOANSHARK:
         LoanShark(To);
         SendClientMessage(To,C_NONE,C_DONE,NULL,NULL);
         break;
      case C_BANK:
         Bank(To);
         SendClientMessage(To,C_NONE,C_DONE,NULL,NULL);
         break;
      case C_GUNSHOP:
         GunShop(To);
         SendClientMessage(To,C_NONE,C_DONE,NULL,NULL);
         break;
      case C_UPDATE:
         if (From==&Noone) {
            ReceivePlayerData(Data,To);
            print_status(To,1); refresh();
         } else {
            DisplaySpyReports(Data,From);
         }
         break;
      case C_NEWNAME:
         clear_line(22); clear_line(23);
         attrset(TextAttr);
         mvaddstr(22,0,"Unfortunately, somebody else is already \
using \"your\" name. Please change it.");
         change_name(1);
         break;
      default:
         sprintf(buf,"%s^%c^%s^%s",From->Name,Code,To->Name,Data);
         mvaddstr(22,0,buf); nice_wait();
         break;
   }
   return 0;
}

void PrepareHighScoreScreen() {
/* Responds to a "starthiscore" message by clearing the screen and */
/* displaying the title for the high scores screen                 */
   attrset(TextAttr);
   clear_screen();
   attrset(TitleAttr);
   mvaddstr(0,29,"H I G H   S C O R E S");
   attrset(TextAttr);
}

void PrintHighScore(char *Data) {
/* Prints a high score coded in "Data"; first word is the index of the  */
/* score (i.e. y screen coordinate), second word is the text, the first */
/* letter of which identifies whether it's to be printed bold or not.   */
   char *cp,wrd[BUFLEN];
   int index;
   cp=Data;
   cp=ExtractWord(wrd,cp);
   index=atoi(wrd);
   if (!cp || strlen(cp)<2) return;
   move(index+2,0);
   attrset(TextAttr);
   if (cp[0]=='B') standout();
   addstr(&cp[1]); 
   if (cp[0]=='B') standend();
}

void PrintMessage(char *text) {
/* Prints a message "text" received via. a "printmessage" message in the */
/* bottom part of the screen.                                            */
   int i,line;
   attrset(TextAttr);
   clear_line(16);
   for (i=0;i<strlen(text);i++) {
      if (text[i]!='^') {
         clear_exceptfor(i+1);
         break;
      }
   }
   line=17; move(line,1);
   for (i=0;i<strlen(text);i++) {
      if (text[i]=='^') {
         line++; move(line,1);
      } else addch(text[i]);
   }
}

void GunShop(struct PLAYER *Play) {
/* Allows player "Play" to buy and sell guns interactively. Passes the */
/* decisions on to the server for sanity checking and implementation.  */
   int i,c,c2;
   char text[BUFLEN],num[80];
   print_status(Play,0);
   clear_bottom();
   for (i=0;i<NUMGUN;i++) {
      sprintf(text,"%c. %-22s %8s",'A'+i,Gun[i].Name,
              FormatPrice(Gun[i].Price,num));
      mvaddstr(17+i/2,(i%2)*40+3,text);
   }
   while (1) {
      strcpy(text,"Will you B>uy, S>ell, or L>eave? ");
      attrset(PromptAttr);
      clear_line(22);
      mvaddstr(22,40-strlen(text)/2,text);
      attrset(TextAttr);
      curs_set(1);
      c=getch(); c=toupper(c);
      curs_set(0);
      if (c=='L') break;
      if (c=='S' || c=='B') {
         clear_line(22);
         if (c=='S' && TotalGunsCarried(Play)==0) {
            mvaddstr(22,10,"You don't have any guns to sell!"); nice_wait();
            clear_line(23);
            continue;
         } else if (c=='B' && TotalGunsCarried(Play)>=Play->Bitches+2) {
            mvaddstr(22,10,"You'll need more bitches to carry any more guns!");
            nice_wait();
            clear_line(23);
            continue;
         }
         attrset(PromptAttr);
         mvaddstr(22,20,"What do you wish to ");
         if (c=='B') addstr("buy? "); else addstr("sell? ");
         curs_set(1);
         attrset(TextAttr);
         c2=getch(); c2=toupper(c2);
         if (c2>='A' && c2<'A'+NUMGUN) {
            c2-='A';
            addstr(Gun[c2].Name);
            if (c=='B') {
               if (Gun[c2].Space > Play->CoatSize) {
                  clear_line(22);
                  mvaddstr(22,10,"You don't have enough space to carry \
that gun!");
                  nice_wait();
                  clear_line(23);
                  continue;
               } else if (Gun[c2].Price > Play->Cash) {
                  clear_line(22);
                  mvaddstr(22,10,"You don't have enough cash to buy that gun!");
                  nice_wait();
                  clear_line(23);
                  continue;
               }
               Play->Cash -= Gun[c2].Price;
               Play->CoatSize -= Gun[c2].Space;
               Play->Carry.Gun[c2]++;
            } else if (c=='S') {
               if (Play->Carry.Gun[c2] == 0) {
                  clear_line(22);
                  mvaddstr(22,10,"You don't have any to sell!");
                  nice_wait(); clear_line(23); continue;
               }
               Play->Cash += Gun[c2].Price;
               Play->CoatSize += Gun[c2].Space;
               Play->Carry.Gun[c2]--;
            }
            print_status(Play,0);
            sprintf(text,"gun^%d^%d",c2,c=='B' ? 1 : -1);
            SendClientMessage(Play,C_NONE,C_BUYOBJECT,NULL,text);
         }
      }
   }
   print_status(Play,1);
}

void LoanShark(struct PLAYER *Play) {
/* Allows player "Play" to pay off loans interactively. */
   char text[BUFLEN];
   long money;
   clear_line(19);
   nice_input("How much money do you give the loan shark? ",text,40,19,1,1,0);
   money=atol(text);
   if (money<0) money=0;
   if (money>Play->Debt) money=Play->Debt;
   if (money>Play->Cash) money=Play->Cash;
   sprintf(text,"%ld",money);
   SendClientMessage(Play,C_NONE,C_PAYLOAN,NULL,text);
}

void Bank(struct PLAYER *Play) {
/* Allows player "Play" to pay in or withdraw money from the bank */
/* interactively.                                                 */
   char text[BUFLEN];
   long money;
   int c;
   attrset(PromptAttr);
   clear_line(18); clear_line(19);
   mvaddstr(18,1,"Do you want to D>eposit in or W>ithdraw from the bank? ");
   c=GetKey("DW");
   nice_input("How much money? ",text,40,19,1,1,0);
   money=atol(text);
   money=labs(money);
   if (c=='W') money=-money;
   sprintf(text,"%ld",money);
   SendClientMessage(Play,C_NONE,C_DEPOSIT,NULL,text);
}
